@@ -2,6 +2,7 @@ import sys, os
 from os import path
 import uuid
 import subprocess
+from StringIO import StringIO
 import re
 from string import Template
 from datetime import datetime, timedelta
@@ -43,26 +44,34 @@ class TarsnapBackend(object):
         self._queried_archives = None
         self._known_archives = []
 
-    def _call(self, *arguments):
+    def call(self, *arguments):
         """
         ``arguments`` is a single list of strings.
         """
         call_with = ['tarsnap']
+        for option in self.options:
+            key = option[0]
+            pre = "-" if len(key) == 1 else "--"
+            call_with.append("%s%s" % (pre, key))
+            for value in option[1:]:
+                call_with.append(value)
         call_with.extend(arguments)
-        for key, value in self.options:
-            call_with.extend(["--%s" % key, value])
-        self.log.debug("Executing: %s" % " ".join(call_with))
-        p = subprocess.Popen(call_with, stdout=subprocess.PIPE,
-                             stderr=subprocess.PIPE)
-        p.wait()
-        if p.returncode != 0:
-            raise TarsnapError('%s' % p.stderr.read())
-        return p.stdout
+        return self._exec_tarsnap(call_with)
 
-    def exec_(self, cmdline):
+    def _exec_tarsnap(self, args):
+        self.log.debug("Executing: %s" % " ".join(args))
+        p = subprocess.Popen(args, stdout=subprocess.PIPE,
+                             stderr=subprocess.PIPE)
+        (stdout, stderr) = p.communicate()
+        if p.returncode != 0:
+            raise TarsnapError('%s' % stderr)
+        return stdout
+
+    def _exec_util(self, cmdline, shell=False):
+        # TODO: can this be merged with _exec_tarsnap into something generic?
         self.log.debug("Executing: %s" % cmdline)
         p = subprocess.Popen(cmdline, shell=True)
-        os.waitpid(p.pid, 0)
+        p.communicate()
         if p.returncode:
             raise RuntimeError('%s failed with exit code %s' % (
                 cmdline, p.returncode))
@@ -81,7 +90,7 @@ class TarsnapBackend(object):
         the first time it is accessed, and then subsequently cached.
         """
         if self._queried_archives is None:
-            response = self._call('--list-archives')
+            response = StringIO(self.call('--list-archives'))
             self._queried_archives = [l.rstrip() for l in response.readlines()]
         return self._queried_archives + self._known_archives
     archives = property(get_archives)
@@ -90,17 +99,42 @@ class TarsnapBackend(object):
         """Return a dict of backups that exist for the given job, by
         parsing the list of archives.
         """
+        # Assemble regular expressions that matche the job's target
+        # filenames, including those based on it's aliases.
         unique = uuid.uuid4().hex
-        target = Template(job.target).substitute({'name': job.name, 'date': unique})
-        regex = re.compile("^%s$" % re.escape(target).replace(unique, '(?P<date>.*?)'))
+        regexes = []
+        for possible_name in [job.name] + (job.aliases or []):
+            target = Template(job.target).substitute(
+                {'name': possible_name, 'date': unique})
+            regexes.append(re.compile("^%s$" %
+                        re.escape(target).replace(unique, '(?P<date>.*?)')))
 
         backups = {}
         for backup_path in self.get_archives():
-            match = regex.match(backup_path)
-            if not match:
+            match = None
+            for regex in regexes:
+                match = regex.match(backup_path)
+                if match:
+                    break
+            else:
+                # Not one of the regexes matched.
                 continue
-            date = parse_date(match.groupdict()['date'], job.dateformat)
-            backups[backup_path] = date
+            try:
+                date = parse_date(match.groupdict()['date'], job.dateformat)
+            except ValueError, e:
+                # This can occasionally happen when multiple archives
+                # share a prefix, say for example you have "windows-$date"
+                # and "windows-data-$date". Since we have to use a generic
+                # .* regex to capture the date part, when processing the
+                # "windows-$date" targets, we'll stumble over entries where
+                # we try to parse "data-$date" as a date. Make sure we
+                # only print a warning, rather than crashing.
+                # TODO: It'd take some work, but we could build a proper
+                # regex based on any given date format string, thus avoiding
+                # the issue for most cases.
+                self.log.error("Ignoring '%s': %s" % (backup_path, e))
+            else:
+                backups[backup_path] = date
 
         return backups
 
@@ -124,7 +158,7 @@ class TarsnapBackend(object):
             if not name in to_keep:
                 self.log.info('Deleting %s' % name)
                 if not self.dryrun:
-                    self._call('-d', '-f', name)
+                    self.call('-d', '-f', name)
                 self.archives.remove(name)
             else:
                 self.log.debug('Keeping %s' % name)
@@ -141,7 +175,11 @@ class TarsnapBackend(object):
             self.log.info('Creating backup: %s' % target)
 
         if not self.dryrun:
-            self._call('-c', '-f', target, *job.sources)
+            args = ['-c']
+            [args.extend(['--exclude', e]) for e in job.excludes]
+            args.extend(['-f', target])
+            args.extend(job.sources)
+            self.call(*args)
         # Add the new backup the list of archives, so we have an up-to-date
         # list without needing to query again.
         self._add_known_archive(target)
@@ -233,6 +271,10 @@ class ExpireCommand(Command):
                             help='only simulate, don\'t delete anything')
 
     def expire(self, job):
+        if not job.deltas:
+            self.log.info(("Skipping '%s', does not define deltas") % job.name)
+            return
+
         self.backend.expire(job)
 
     def run(self, job):
@@ -270,17 +312,22 @@ class MakeCommand(ExpireCommand):
                                 'using --sources')
 
     def run(self, job):
+        if not job.sources:
+            self.log.info(("Skipping '%s', does not define sources") % job.name)
+            return
+
         # Determine whether we can run this job. If any of the sources
         # are missing, or any source directory is empty, we skip this job.
         sources_missing = False
-        for source in job.sources:
-            if not path.exists(source):
-                sources_missing = True
-                break
-            if path.isdir(source) and not os.listdir(source):
-                # directory is empty
-                sources_missing = True
-                break
+        if not job.force:
+            for source in job.sources:
+                if not path.exists(source):
+                    sources_missing = True
+                    break
+                if path.isdir(source) and not os.listdir(source):
+                    # directory is empty
+                    sources_missing = True
+                    break
 
         # Do a new backup
         skipped = False
@@ -295,12 +342,12 @@ class MakeCommand(ExpireCommand):
             skipped = True
         else:
             if job.exec_before:
-                self.backend.exec_(job.exec_before)
+                self.backend._exec_util(job.exec_before)
             try:
                 self.backend.make(job)
             finally:
                 if job.exec_after:
-                    self.backend.exec_(job.exec_after)
+                    self.backend._exec_util(job.exec_after)
 
         # Expire old backups, but only bother if either we made a new
         # backup, or if expire was explicitly requested.
@@ -323,9 +370,11 @@ def parse_args(argv):
     group = parser.add_mutually_exclusive_group()
     group.add_argument('-q', action='store_true', dest='quiet', help='be quiet')
     group.add_argument('-v', action='store_true', dest='verbose', help='be verbose')
-    parser.add_argument('-o', metavar=('name', 'value'), nargs=2,
+    # We really want nargs=(1,2), but since this isn't available, we can
+    # just asl well support an arbitrary number of values for each -o.
+    parser.add_argument('-o', metavar=('name', 'value'), nargs='+',
                         dest='tarsnap_options', default=[], action='append',
-                        help='option to pass to tarsnap')
+                        help='option to pass to tarsnap',)
     parser.add_argument('--config', '-c', help='use the given config file')
 
     group = parser.add_argument_group(
@@ -424,7 +473,7 @@ def main(argv):
     else:
         # Only a single job, as given on the command line
         jobs = {None: Job(**{'target': args.target, 'dateformat': args.dateformat,
-                       '      deltas': args.deltas, 'sources': args.sources})}
+                             'deltas': args.deltas, 'sources': args.sources})}
 
     # Validate the requested list of jobs to run
     if args.jobs:
